from typing import Callable
import importlib
import functools
from wax.lessweb.utils import func_arg_spec
from pathlib import Path


class Importer:
    def __call__(self, package: str):
        return importlib.import_module(package)

    def __getattr__(self, item):
        return importlib.import_module(item)


lib = Importer()


def eval_func(lambda_func: str, **inject) -> Callable:
    kwargs = {}
    fn = eval(lambda_func)
    for arg_name in func_arg_spec(fn):
        if arg_name in inject:
            kwargs[arg_name] = inject[arg_name]
    return functools.partial(fn, **kwargs)


lib_path = Path('script')


def default_func(func_name: str, **inject) -> Callable:
    assert func_name.startswith('@')
    lambda_func = (lib_path / f'{func_name[1:]}.py').read_text(encoding='utf-8').strip()
    return eval_func(lambda_func, **inject)


def is_evalable(obj):
    return isinstance(obj, list) and obj and \
           isinstance(obj[0], str) and obj[0].startswith('@')


def deep_eval(obj, **inject):
    if not is_evalable(obj):
        if isinstance(obj, dict):
            return {key: deep_eval(val) for key, val in obj.items()}
        elif isinstance(obj, list):
            return [deep_eval(item) for item in obj]
        else:
            return obj
    fn = default_func(obj[0], **inject)
    args = [deep_eval(arg) for arg in obj[1:]]
    return fn(*args)



from typing import get_type_hints, TypeVar, Type


__all__ = ["Storage"]

T = TypeVar('T')


class Storage(dict):
    """
    A Storage object is like a dictionary except `obj.foo` can be used
    in addition to `obj['foo']`.

        >>> o = Storage(a=1)
        >>> o.a
        1
        >>> o['a']
        1
        >>> o.a = 2
        >>> o['a']
        2
        >>> del o.a
        >>> o.a
        Traceback (most recent call last):
            ...
        AttributeError: 'a'
    """
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as k:
            raise AttributeError(k)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError as k:
            raise AttributeError(k)

    def __repr__(self):
        return '<Storage ' + dict.__repr__(self) + '>'

    def __sub__(self, other):
        if isinstance(other, str):
            if other in self:
                del self[other]
        else:
            for key in other:
                self.__sub__(key)
        return self

    @staticmethod
    def type_hints(cls: Type) -> 'Storage':
        result = Storage()
        for prop_name, prop_type in get_type_hints(cls).items():
            if prop_name[0] != '_':
                result[prop_name] = prop_type
            else:
                pub_name = prop_name[1:]
                if isinstance(getattr(cls, pub_name, None), property):
                    result[pub_name] = prop_type
        return result

    @staticmethod
    def of(obj) -> 'Storage':
        result = Storage()
        for name in get_type_hints(obj).keys():
            if name[0] == '_':
                name = name[1:]
            if hasattr(obj, name):
                result[name] = getattr(obj, name)
        return result

    def to(self, cls: Type[T]) -> T:
        obj = cls()
        for key, val in self.items():
            setattr(obj, key, val)
        return obj



from wax.lessweb import Application, Context
from wax.lessweb.plugin.redisplugin import RedisPlugin
from wax.load_config import config


def allow_cors(ctx: Context):
    ctx.response.send_access_allow()
    return ctx()


app = Application()
app.add_plugin(RedisPlugin(**config['redis']))

app.add_interceptor('.*', method='*', dealer=allow_cors)
app.add_options_mapping('.*', lambda:'')

from wax.mock_api import mock_dealer, SwaggerData
from wax.tag_api import make_kotlin_code, make_solution_list, make_openapi_json
wax_api_prefix = config['mockapi-prefix']
app.add_mapping(f'{wax_api_prefix}/.*', method='*', dealer=mock_dealer)
app.add_get_mapping('/openapi.json', dealer=make_openapi_json)
app.add_get_mapping('/openapi.kt', dealer=make_kotlin_code)
app.add_get_mapping('/solution.md', dealer=make_solution_list)

from wax.tag_api import operation_list, operation_example, operation_detail, operation_edit_state, compare_swagger
app.add_get_mapping('/', operation_list)
app.add_get_mapping('/tag/{tag}', operation_list)
app.add_post_mapping('/op/state', operation_edit_state)
app.add_post_mapping('/op/diff', compare_swagger)
app.add_get_mapping('/op/{opId}', operation_detail)
app.add_get_mapping('/op/{opId}/example', operation_example)



from typing import Dict, Any
import json
import jsonschema
import itertools
import datetime
from pathlib import Path
from wax.load_config import config
from wax.pack_util import packed, dir_mtime
from wax.jsonschema_util import jsonschema_to_rows


class SwaggerData:
    json_path = ''
    redis_prefix = 'waxapi::'
    swagger_data: Dict = {}
    resolver = None
    last_modify = 0

    @classmethod
    def init(cls, json_path):
        title, version = config['title'], config['version']
        if not title:
            print('Title cannot be empty, please edit config.json!')
            exit(1)
        if not version:
            print('Version cannot be empty, please edit config.json!')
            exit(1)
        if not Path(json_path).is_dir():
            print(f'{json_path} is not exist or is not directory!')
        cls.json_path = json_path
        cls.last_modify = dir_mtime(json_path)
        cls.swagger_data = packed(json_path, title, version)
        cls.resolver = jsonschema.RefResolver.from_schema(cls.swagger_data)
        cls.redis_prefix = 'waxapi::' + title + '::' + version + '::'

    @classmethod
    def get(cls) -> Dict:
        last_modify = dir_mtime(cls.json_path)
        if last_modify != cls.last_modify:
            title, version = config['title'], config['version']
            cls.swagger_data = packed(cls.json_path, title, version)
            cls.resolver = jsonschema.RefResolver.from_schema(cls.swagger_data)
            cls.last_modify = last_modify
            print('Reloaded at %s.' % datetime.datetime.fromtimestamp(int(last_modify)))
        return cls.swagger_data


def parse_operation(swagger_data, endpoint:Dict, method:str) -> Dict:
    """
    :return {'params': ..., 'requests': ..., 'responses': ...}
    """
    operation = endpoint[method.lower()]
    data = {'params': {}, 'requests': [], 'responses': []}
    params = {'path': [], 'query': [], 'header': []}
    for param in itertools.chain(endpoint.get('parameters', []), operation.get('parameters', [])):
        for source in params.keys():
            if param.get('in', '') == source:
                params[source].append(param)
    data['params'].update(params)
    for status_code, response_val in operation.get('responses', {}).items():
        for content_key, content_val in response_val.get('content', {}).items():
            data['responses'].append({
                'status_code': status_code,
                'content_type': content_key,
                'rows': jsonschema_to_rows('', '+', content_val.get('schema', {}), swagger_data, required=[]),
                'examples': {key: json.dumps(val['value'], ensure_ascii=False, indent=4)
                             for key, val in content_val.get('examples', {}).items()}
            })
    for content_key, content_val in operation.get('requestBody', {}).get('content', {}).items():
        data['requests'].append({
            'content_type': content_key,
            'rows': jsonschema_to_rows('', '+', content_val.get('schema', {}), swagger_data, required=[])
        })
    return data



import sys
from typing import Any, Tuple, Dict, Type, get_type_hints, Callable
from contextlib import contextmanager
import json
from pathlib import Path
import pickle
import re
import inspect
from inspect import Signature


__all__ = ["eafp", "_nil", "re_standardize", "func_arg_spec", "makedir"]
NEW_INSPECT = sys.version_info[:3] >= (3, 8, 0)


def eafp(ask: Callable, default: Any) -> Any:
    """
    Easier to ask for forgiveness than permission
    `x = eafp(lambda: int('a'), 0)` is equivalent to `x = int('a') ?? 0`
    """
    try:
        return ask()
    except:
        return default


class Nil:
    def __init__(self, value):
        self.value = value

    def __bool__(self):
        return False

    def __eq__(self, other):
        return isinstance(other, Nil) and self.value == other.value


_nil = Nil(0)


def re_standardize(pattern: str) -> str:
    """
        >>> pattern = re_standardize('/add/{x}/{y}')
        >>> pattern
        '^/add/(?P<x>[^/]+)/(?P<y>[^/]+)$'
        >>> re.search(pattern, '/add/234/5').groupdict()
        {'x': '234', 'y': '5'}
        >>> re.search(pattern, '/add//add') is None
        True
        >>> re.search(pattern, '/add/1/2/') is None
        True

    """
    if not pattern:
        return '^$'
    if pattern[0] != '^':
        pattern = '^' + pattern
    if pattern[-1] != '$':
        pattern = pattern + '$'
    def _repl(obj):
        x = obj.groups()[0]
        return '(?P<%s>[^/]+)' % x

    return re.sub(r'\{([^0-9].*?)\}', _repl, pattern)


def func_arg_spec(fn: Any) -> Dict[str, Tuple[Type, bool, bool]]:
    empty = Signature.empty
    arg_spec = {}  # name: (type_, has_default, positional-only)
    for name, param in inspect.signature(fn).parameters.items():
        arg_spec[name] = (
            Any if param.annotation is empty else param.annotation,
            param.default is not empty,
            NEW_INSPECT and param.kind == 0
        )
    return arg_spec


def makedir(real_path: str) -> None:
    Path(real_path).mkdir(parents=True, exist_ok=True)
